@@ -16,7 +16,7 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 from __future__ import absolute_import
 
-VERSION = (1, 0, 11)
+VERSION = (1, 3, 0)
 __version__ = '.'.join(map(str, VERSION[0:3])) + ''.join(VERSION[3:])
 __author__ = 'Barry Pederson'
 __maintainer__ = 'Ask Solem'
@@ -29,22 +29,42 @@ __docformat__ = 'restructuredtext'
 #
 # Pull in the public items from the various sub-modules
 #
-from .basic_message import Message
-from .channel import Channel
-from .connection import Connection
-from .exceptions import (
+from .basic_message import Message      # noqa
+from .channel import Channel            # noqa
+from .connection import Connection      # noqa
+from .exceptions import (               # noqa
     AMQPError,
     ConnectionError,
+    RecoverableConnectionError,
+    IrrecoverableConnectionError,
     ChannelError,
-    ConsumerCancel,
+    RecoverableChannelError,
+    IrrecoverableChannelError,
+    ConsumerCancelled,
+    ContentTooLarge,
+    NoConsumers,
+    ConnectionForced,
+    InvalidPath,
+    AccessRefused,
+    NotFound,
+    ResourceLocked,
+    PreconditionFailed,
+    FrameError,
+    FrameSyntaxError,
+    InvalidCommand,
+    ChannelNotOpen,
+    UnexpectedFrame,
+    ResourceError,
+    NotAllowed,
+    AMQPNotImplementedError,
+    InternalError,
+    error_for_code,
+    __all__ as _all_exceptions,
 )
+from .utils import promise  # noqa
 
 __all__ = [
     'Connection',
     'Channel',
     'Message',
-    'AMQPError',
-    'ConnectionError',
-    'ChannelError',
-    'ConsumerCancel',
-]
+] + _all_exceptions
