@@ -19,11 +19,12 @@ from __future__ import absolute_import
 import logging
 
 from collections import defaultdict
-from Queue import Queue
 from warnings import warn
 
 from .abstract_channel import AbstractChannel
-from .exceptions import ChannelError, ConsumerCancel
+from .exceptions import ChannelError, ConsumerCancelled, error_for_code
+from .five import Queue
+from .protocol import basic_return_t, queue_declare_ok_t
 from .serialization import AMQPWriter
 
 __all__ = ['Channel']
@@ -62,8 +63,11 @@ class Channel(AbstractChannel):
         is left as plain bytes.
 
         """
-        if channel_id is None:
+        if channel_id:
+            connection._claim_channel_id(channel_id)
+        else:
             channel_id = connection._get_free_channel_id()
+
         AMQP_LOGGER.debug('using channel_id: %d', channel_id)
 
         super(Channel, self).__init__(connection, channel_id)
@@ -74,9 +78,15 @@ class Channel(AbstractChannel):
         self.callbacks = {}
         self.cancel_callbacks = {}
         self.auto_decode = auto_decode
-        self.events = defaultdict(list)
+        self.events = defaultdict(set)
         self.no_ack_consumers = set()
 
+        # set first time basic_publish_confirm is called
+        # and publisher confirms are enabled for this channel.
+        self._confirm_selected = False
+        if self.connection.confirm_publish:
+            self.basic_publish = self.basic_publish_confirm
+
         self._x_open()
 
     def _do_close(self):
@@ -84,14 +94,20 @@ class Channel(AbstractChannel):
         with the server."""
         AMQP_LOGGER.debug('Closed channel #%d', self.channel_id)
         self.is_open = False
-        if self.connection:
-            self.connection.channels.pop(self.channel_id, None)
-        self.channel_id = self.connection = None
+        channel_id, self.channel_id = self.channel_id, None
+        connection, self.connection = self.connection, None
+        if connection:
+            connection.channels.pop(channel_id, None)
+            connection._avail_channel_ids.append(channel_id)
         self.callbacks.clear()
         self.cancel_callbacks.clear()
         self.events.clear()
         self.no_ack_consumers.clear()
 
+    def _do_revive(self):
+        self.is_open = False
+        self._x_open()
+
     def close(self, reply_code=0, reply_text='', method_sig=(0, 0)):
         """Request a channel close
 
@@ -140,7 +156,7 @@ class Channel(AbstractChannel):
 
         """
         try:
-            if not self.is_open:
+            if not self.is_open or self.connection is None:
                 return
 
             args = AMQPWriter()
@@ -203,15 +219,18 @@ class Channel(AbstractChannel):
                 is the ID of the method.
 
         """
+
         reply_code = args.read_short()
         reply_text = args.read_shortstr()
         class_id = args.read_short()
         method_id = args.read_short()
 
         self._send_method((20, 41))
-        self._do_close()
+        self._do_revive()
 
-        raise ChannelError(reply_code, reply_text, (class_id, method_id))
+        raise error_for_code(
+            reply_code, reply_text, (class_id, method_id), ChannelError,
+        )
 
     def _close_ok(self, args):
         """Confirm a channel close
@@ -449,7 +468,7 @@ class Channel(AbstractChannel):
     #
 
     def exchange_declare(self, exchange, type, passive=False, durable=False,
-            auto_delete=True, nowait=False, arguments=None):
+                         auto_delete=True, nowait=False, arguments=None):
         """Declare exchange, create if needed
 
         This method creates an exchange if it does not already exist,
@@ -602,8 +621,7 @@ class Channel(AbstractChannel):
         """
         pass
 
-    def exchange_delete(self, exchange, if_unused=False,
-            nowait=False):
+    def exchange_delete(self, exchange, if_unused=False, nowait=False):
         """Delete an exchange
 
         This method deletes an exchange.  When an exchange is deleted
@@ -667,7 +685,7 @@ class Channel(AbstractChannel):
         pass
 
     def exchange_bind(self, destination, source='', routing_key='',
-            nowait=False, arguments=None):
+                      nowait=False, arguments=None):
         """This method binds an exchange to an exchange.
 
         RULE:
@@ -755,7 +773,7 @@ class Channel(AbstractChannel):
             ])
 
     def exchange_unbind(self, destination, source='', routing_key='',
-            nowait=False, arguments=None):
+                        nowait=False, arguments=None):
         """This method unbinds an exchange from an exchange.
 
         RULE:
@@ -864,7 +882,7 @@ class Channel(AbstractChannel):
     #
 
     def queue_bind(self, queue, exchange='', routing_key='',
-            nowait=False, arguments=None):
+                   nowait=False, arguments=None):
         """Bind queue to an exchange
 
         This method binds a queue to an exchange.  Until a queue is
@@ -986,7 +1004,7 @@ class Channel(AbstractChannel):
         pass
 
     def queue_unbind(self, queue, exchange, routing_key='',
-            nowait=False, arguments=None):
+                     nowait=False, arguments=None):
         """Unbind a queue from an exchange
 
         This method unbinds a queue from an exchange.
@@ -1061,8 +1079,8 @@ class Channel(AbstractChannel):
         pass
 
     def queue_declare(self, queue='', passive=False, durable=False,
-            exclusive=False, auto_delete=True, nowait=False,
-            arguments=None):
+                      exclusive=False, auto_delete=True, nowait=False,
+                      arguments=None):
         """Declare queue, create if needed
 
         This method creates or checks a queue.  When creating a new
@@ -1261,13 +1279,14 @@ class Channel(AbstractChannel):
                 this count.
 
         """
-        queue = args.read_shortstr()
-        message_count = args.read_long()
-        consumer_count = args.read_long()
-        return queue, message_count, consumer_count
+        return queue_declare_ok_t(
+            args.read_shortstr(),
+            args.read_long(),
+            args.read_long(),
+        )
 
-    def queue_delete(self, queue='', if_unused=False, if_empty=False,
-            nowait=False):
+    def queue_delete(self, queue='',
+                     if_unused=False, if_empty=False, nowait=False):
         """Delete a queue
 
         This method deletes a queue.  When a queue is deleted any
@@ -1595,14 +1614,15 @@ class Channel(AbstractChannel):
                 channel or connection exception.
 
         """
-        self.no_ack_consumers.discard(consumer_tag)
-        args = AMQPWriter()
-        args.write_shortstr(consumer_tag)
-        args.write_bit(nowait)
-        self._send_method((60, 30), args)
-        return self.wait(allowed_methods=[
-            (60, 31),  # Channel.basic_cancel_ok
-        ])
+        if self.connection is not None:
+            self.no_ack_consumers.discard(consumer_tag)
+            args = AMQPWriter()
+            args.write_shortstr(consumer_tag)
+            args.write_bit(nowait)
+            self._send_method((60, 30), args)
+            return self.wait(allowed_methods=[
+                (60, 31),  # Channel.basic_cancel_ok
+            ])
 
     def _basic_cancel_notify(self, args):
         """Consumer cancelled by server.
@@ -1615,7 +1635,7 @@ class Channel(AbstractChannel):
         if callback:
             callback(consumer_tag)
         else:
-            raise ConsumerCancel('tag %r' % (consumer_tag, ))
+            raise ConsumerCancelled(consumer_tag, (60, 30))
 
     def _basic_cancel_ok(self, args):
         """Confirm a cancelled consumer
@@ -1646,8 +1666,8 @@ class Channel(AbstractChannel):
         return self.cancel_callbacks.pop(consumer_tag, None)
 
     def basic_consume(self, queue='', consumer_tag='', no_local=False,
-            no_ack=False, exclusive=False, nowait=False,
-            callback=None, arguments=None, on_cancel=None):
+                      no_ack=False, exclusive=False, nowait=False,
+                      callback=None, arguments=None, on_cancel=None):
         """Start a queue consumer
 
         This method asks the server to start a "consumer", which is a
@@ -2012,8 +2032,8 @@ class Channel(AbstractChannel):
         }
         return msg
 
-    def basic_publish(self, msg, exchange='', routing_key='',
-            mandatory=False, immediate=False):
+    def _basic_publish(self, msg, exchange='', routing_key='',
+                       mandatory=False, immediate=False):
         """Publish a message
 
         This method publishes a message to a specific exchange. The
@@ -2087,6 +2107,15 @@ class Channel(AbstractChannel):
         args.write_bit(immediate)
 
         self._send_method((60, 40), args, msg)
+    basic_publish = _basic_publish
+
+    def basic_publish_confirm(self, *args, **kwargs):
+        if not self._confirm_selected:
+            self._confirm_selected = True
+            self.confirm_select()
+        ret = self._basic_publish(*args, **kwargs)
+        self.wait([(60, 80)])
+        return ret
 
     def basic_qos(self, prefetch_size, prefetch_count, a_global):
         """Specify quality of service
@@ -2322,14 +2351,13 @@ class Channel(AbstractChannel):
                 message was published.
 
         """
-        reply_code = args.read_short()
-        reply_text = args.read_shortstr()
-        exchange = args.read_shortstr()
-        routing_key = args.read_shortstr()
-
-        self.returned_messages.put(
-            (reply_code, reply_text, exchange, routing_key, msg)
-        )
+        self.returned_messages.put(basic_return_t(
+            args.read_short(),
+            args.read_shortstr(),
+            args.read_shortstr(),
+            args.read_shortstr(),
+            msg,
+        ))
 
     #############
     #
@@ -2457,7 +2485,7 @@ class Channel(AbstractChannel):
     def _basic_ack_recv(self, args):
         delivery_tag = args.read_longlong()
         multiple = args.read_bit()
-        for callback in self.handlers['basic_ack']:
+        for callback in self.events['basic_ack']:
             callback(delivery_tag, multiple)
 
     _METHOD_MAP = {
@@ -2489,7 +2517,7 @@ class Channel(AbstractChannel):
         (90, 11): _tx_select_ok,
         (90, 21): _tx_commit_ok,
         (90, 31): _tx_rollback_ok,
-        }
+    }
 
     _IMMEDIATE_METHODS = [
         (60, 50),  # basic_return
