@@ -18,6 +18,8 @@ from __future__ import absolute_import
 
 import logging
 import socket
+
+from array import array
 try:
     from ssl import SSLError
 except ImportError:
@@ -27,7 +29,12 @@ except ImportError:
 from . import __version__
 from .abstract_channel import AbstractChannel
 from .channel import Channel
-from .exceptions import AMQPError, ChannelError, ConnectionError
+from .exceptions import (
+    AMQPNotImplementedError, ChannelError, ResourceError,
+    ConnectionForced, ConnectionError, error_for_code,
+    RecoverableConnectionError, RecoverableChannelError,
+)
+from .five import items, range, values
 from .method_framing import MethodReader, MethodWriter
 from .serialization import AMQPWriter
 from .transport import create_transport
@@ -77,21 +84,12 @@ class Connection(AbstractChannel):
     prev_recv = None
     missed_heartbeats = 0
 
-    def __init__(self,
-        host='localhost',
-        userid='guest',
-        password='guest',
-        login_method='AMQPLAIN',
-        login_response=None,
-        virtual_host='/',
-        locale='en_US',
-        client_properties=None,
-        ssl=False,
-        connect_timeout=None,
-        channel_max=None,
-        frame_max=None,
-        heartbeat=0,
-        **kwargs):
+    def __init__(self, host='localhost', userid='guest', password='guest',
+                 login_method='AMQPLAIN', login_response=None,
+                 virtual_host='/', locale='en_US', client_properties=None,
+                 ssl=False, connect_timeout=None, channel_max=None,
+                 frame_max=None, heartbeat=0, on_blocked=None,
+                 on_unblocked=None, confirm_publish=False, **kwargs):
         """Create a connection to the specified host, which should be
         a 'host[:port]', such as 'localhost', or '1.2.3.4:5672'
         (defaults to 'localhost', if a port is not specified then
@@ -129,6 +127,14 @@ class Connection(AbstractChannel):
         self.frame_max = frame_max
         self.heartbeat = heartbeat
 
+        self.confirm_publish = confirm_publish
+
+        # Callbacks
+        self.on_blocked = on_blocked
+        self.on_unblocked = on_unblocked
+
+        self._avail_channel_ids = array('H', range(self.channel_max, 0, -1))
+
         # Properties set in the Start method
         self.version_major = 0
         self.version_minor = 0
@@ -163,7 +169,7 @@ class Connection(AbstractChannel):
         try:
             self.transport.close()
 
-            temp_list = [x for x in self.channels.values() if x is not self]
+            temp_list = [x for x in values(self.channels) if x is not self]
             for ch in temp_list:
                 ch._do_close()
         except socket.error:
@@ -172,12 +178,19 @@ class Connection(AbstractChannel):
             self.transport = self.connection = self.channels = None
 
     def _get_free_channel_id(self):
-        for i in xrange(1, self.channel_max + 1):
-            if i not in self.channels:
-                return i
-        raise AMQPError(
-            'No free channel ids, current=%d, channel_max=%d' % (
-                len(self.channels), self.channel_max))
+        try:
+            return self._avail_channel_ids.pop()
+        except IndexError:
+            raise ResourceError(
+                'No free channel ids, current={0}, channel_max={1}'.format(
+                    len(self.channels), self.channel_max), (20, 10))
+
+    def _claim_channel_id(self, channel_id):
+        try:
+            return self._avail_channel_ids.remove(channel_id)
+        except ValueError:
+            raise ConnectionError(
+                'Channel %r already open' % (channel_id, ))
 
     def _wait_method(self, channel_id, allowed_methods):
         """Wait for a method from the server destined for
@@ -202,10 +215,10 @@ class Connection(AbstractChannel):
             channel, method_sig, args, content = \
                 self.method_reader.read_method()
 
-            if (channel == channel_id) \
-                    and ((allowed_methods is None) \
-                    or (method_sig in allowed_methods) \
-                    or (method_sig == (20, 40))):
+            if channel == channel_id and (
+                    allowed_methods is None or
+                    method_sig in allowed_methods or
+                    method_sig == (20, 40)):
                 return method_sig, args, content
 
             #
@@ -215,7 +228,8 @@ class Connection(AbstractChannel):
             #
             if channel and method_sig in self.Channel._IMMEDIATE_METHODS:
                 self.channels[channel].dispatch_method(
-                        method_sig, args, content)
+                    method_sig, args, content,
+                )
                 continue
 
             #
@@ -223,7 +237,7 @@ class Connection(AbstractChannel):
             # this method for later
             #
             self.channels[channel].method_queue.append(
-                    (method_sig, args, content)
+                (method_sig, args, content),
             )
 
             #
@@ -261,23 +275,25 @@ class Connection(AbstractChannel):
         """Wait for an event on a channel."""
         chanmap = self.channels
         chanid, method_sig, args, content = self._wait_multiple(
-                chanmap, None, timeout=timeout)
+            chanmap, None, timeout=timeout,
+        )
 
         channel = chanmap[chanid]
 
-        if content \
-        and channel.auto_decode \
-        and hasattr(content, 'content_encoding'):
+        if (content and
+                channel.auto_decode and
+                hasattr(content, 'content_encoding')):
             try:
                 content.body = content.body.decode(content.content_encoding)
             except Exception:
                 pass
 
-        amqp_method = self._method_override.get(method_sig) or \
-                        channel._METHOD_MAP.get(method_sig, None)
+        amqp_method = (self._method_override.get(method_sig) or
+                       channel._METHOD_MAP.get(method_sig, None))
 
         if amqp_method is None:
-            raise Exception('Unknown AMQP method (%d, %d)' % method_sig)
+            raise AMQPNotImplementedError(
+                'Unknown AMQP method {0!r}'.format(method_sig))
 
         if content is None:
             return amqp_method(channel, args)
@@ -294,23 +310,26 @@ class Connection(AbstractChannel):
         try:
             try:
                 return self.method_reader.read_method()
-            except SSLError, exc:
+            except SSLError as exc:
                 # http://bugs.python.org/issue10272
                 if 'timed out' in str(exc):
                     raise socket.timeout()
+               # Non-blocking SSL sockets can throw SSLError
+                if 'The operation did not complete' in str(exc):
+                    raise socket.timeout()
                 raise
         finally:
             if prev != timeout:
                 sock.settimeout(prev)
 
     def _wait_multiple(self, channels, allowed_methods, timeout=None):
-        for channel_id, channel in channels.iteritems():
+        for channel_id, channel in items(channels):
             method_queue = channel.method_queue
             for queued_method in method_queue:
                 method_sig = queued_method[0]
-                if (allowed_methods is None) \
-                or (method_sig in allowed_methods) \
-                or (method_sig == (20, 40)):
+                if (allowed_methods is None or
+                        method_sig in allowed_methods or
+                        method_sig == (20, 40)):
                     method_queue.remove(queued_method)
                     method_sig, args, content = queued_method
                     return channel_id, method_sig, args, content
@@ -321,10 +340,10 @@ class Connection(AbstractChannel):
         while 1:
             channel, method_sig, args, content = read_timeout(timeout)
 
-            if channel in channels and \
-                    (allowed_methods is None or \
-                            method_sig in allowed_methods or \
-                            method_sig == (20, 40)):
+            if channel in channels and (
+                    allowed_methods is None or
+                    method_sig in allowed_methods or
+                    method_sig == (20, 40)):
                 return channel, method_sig, args, content
 
             # Not the channel and/or method we were looking for. Queue
@@ -345,7 +364,7 @@ class Connection(AbstractChannel):
         exchange = args.read_shortstr()
         routing_key = args.read_shortstr()
 
-        exc = ChannelError('basic.return', reply_code, reply_text, (50, 60))
+        exc = error_for_code(reply_code, reply_text, (50, 60), ChannelError)
         handlers = channel.events.get('basic_return')
         if not handlers:
             raise exc
@@ -482,7 +501,18 @@ class Connection(AbstractChannel):
 
         self._x_close_ok()
 
-        raise ConnectionError(reply_code, reply_text, (class_id, method_id))
+        raise error_for_code(reply_code, reply_text,
+                             (class_id, method_id), ConnectionError)
+
+    def _blocked(self, args):
+        """RabbitMQ Extension."""
+        reason = args.read_shortstr()
+        if self.on_blocked:
+            return self.on_blocked(reason)
+
+    def _unblocked(self, *args):
+        if self.on_unblocked:
+            return self.on_unblocked()
 
     def _x_close_ok(self):
         """Confirm a connection close
@@ -694,9 +724,11 @@ class Connection(AbstractChannel):
         self.mechanisms = args.read_longstr().split(' ')
         self.locales = args.read_longstr().split(' ')
 
-        AMQP_LOGGER.debug(START_DEBUG_FMT,
+        AMQP_LOGGER.debug(
+            START_DEBUG_FMT,
             self.version_major, self.version_minor,
-            self.server_properties, self.mechanisms, self.locales)
+            self.server_properties, self.mechanisms, self.locales,
+        )
 
     def _x_start_ok(self, client_properties, mechanism, response, locale):
         """Select security mechanism and locale
@@ -751,6 +783,10 @@ class Connection(AbstractChannel):
             if 'capabilities' not in client_properties:
                 client_properties['capabilities'] = {}
             client_properties['capabilities']['consumer_cancel_notify'] = True
+        if self.server_capabilities.get('connection.blocked'):
+            if 'capabilities' not in client_properties:
+                client_properties['capabilities'] = {}
+            client_properties['capabilities']['connection.blocked'] = True
         args = AMQPWriter()
         args.write_table(client_properties)
         args.write_shortstr(mechanism)
@@ -833,7 +869,7 @@ class Connection(AbstractChannel):
         self.prev_sent, self.prev_recv = sent_now, recv_now
 
         if self.missed_heartbeats >= rate:
-            raise ConnectionError('Too many heartbeats missed')
+            raise ConnectionForced('Too many heartbeats missed')
 
     def _x_tune_ok(self, channel_max, frame_max, heartbeat):
         """Negotiate connection tuning parameters
@@ -906,12 +942,24 @@ class Connection(AbstractChannel):
         (10, 41): _open_ok,
         (10, 50): _close,
         (10, 51): _close_ok,
+        (10, 60): _blocked,
+        (10, 61): _unblocked,
     }
 
     _IMMEDIATE_METHODS = []
-    connection_errors = (ConnectionError,
-                         socket.error,
-                         IOError,
-                         OSError,
-                         AttributeError)
+    connection_errors = (
+        ConnectionError,
+        socket.error,
+        IOError,
+        OSError,
+    )
     channel_errors = (ChannelError, )
+    recoverable_connection_errors = (
+        RecoverableConnectionError,
+        socket.error,
+        IOError,
+        OSError,
+    )
+    recoverable_channel_errors = (
+        RecoverableChannelError,
+    )
