@@ -48,10 +48,13 @@ except:
 
 from struct import pack, unpack
 
-from .exceptions import AMQPError
+from .exceptions import UnexpectedFrame
+from .utils import set_cloexec
 
 AMQP_PORT = 5672
 
+EMPTY_BUFFER = bytes()
+
 # Yes, Advanced Message Queuing Protocol Protocol is redundant
 AMQP_PROTOCOL_HEADER = 'AMQP\x01\x01\x00\x09'.encode('latin_1')
 
@@ -63,7 +66,7 @@ class _AbstractTransport(object):
     """Common superclass for TCP and SSL transports"""
 
     def __init__(self, host, connect_timeout):
-        msg = 'socket.getaddrinfo() for %s returned an empty list' % host
+        msg = None
         port = AMQP_PORT
 
         m = IPV6_LITERAL.match(host)
@@ -83,9 +86,14 @@ class _AbstractTransport(object):
             af, socktype, proto, canonname, sa = res
             try:
                 self.sock = socket.socket(af, socktype, proto)
+                try:
+                    set_cloexec(self.sock, True)
+                except NotImplementedError:
+                    pass
                 self.sock.settimeout(connect_timeout)
                 self.sock.connect(sa)
-            except socket.error, msg:
+            except socket.error as exc:
+                msg = exc
                 self.sock.close()
                 self.sock = None
                 last_err = msg
@@ -139,23 +147,23 @@ class _AbstractTransport(object):
             self.sock.close()
             self.sock = None
 
-    def read_frame(self):
-        """Read an AMQP frame."""
-        frame_type, channel, size = unpack('>BHI', self._read(7, True))
-        payload = self._read(size)
-        ch = ord(self._read(1))
+    def read_frame(self, unpack=unpack):
+        read = self._read
+        frame_type, channel, size = unpack('>BHI', read(7, True))
+        payload = read(size)
+        ch = ord(read(1))
         if ch == 206:  # '\xce'
             return frame_type, channel, payload
         else:
-            raise AMQPError(
-                'Framing Error, received 0x%02x while expecting 0xce' % ch)
+            raise UnexpectedFrame(
+                'Received 0x{0:02x} while expecting 0xce'.format(ch))
 
     def write_frame(self, frame_type, channel, payload):
-        """Write out an AMQP frame."""
         size = len(payload)
-        self._write(
-            pack('>BHI%dsB' % size, frame_type, channel, size, payload, 0xce),
-        )
+        self._write(pack(
+            '>BHI%dsB' % size,
+            frame_type, channel, size, payload, 0xce,
+        ))
 
 
 class SSLTransport(_AbstractTransport):
@@ -164,7 +172,7 @@ class SSLTransport(_AbstractTransport):
     def __init__(self, host, connect_timeout, ssl):
         if isinstance(ssl, dict):
             self.sslopts = ssl
-        self.sslobj = None
+        self._read_buffer = EMPTY_BUFFER
         super(SSLTransport, self).__init__(host, connect_timeout)
 
     def _setup_transport(self):
@@ -173,43 +181,51 @@ class SSLTransport(_AbstractTransport):
         lower version."""
         if HAVE_PY26_SSL:
             if hasattr(self, 'sslopts'):
-                self.sslobj = ssl.wrap_socket(self.sock, **self.sslopts)
+                self.sock = ssl.wrap_socket(self.sock, **self.sslopts)
             else:
-                self.sslobj = ssl.wrap_socket(self.sock)
-            self.sslobj.do_handshake()
+                self.sock = ssl.wrap_socket(self.sock)
+            self.sock.do_handshake()
         else:
-            self.sslobj = socket.ssl(self.sock)
+            self.sock = socket.ssl(self.sock)
+        self._quick_recv = self.sock.read
 
     def _shutdown_transport(self):
         """Unwrap a Python 2.6 SSL socket, so we can call shutdown()"""
-        if HAVE_PY26_SSL and (self.sslobj is not None):
-            self.sock = self.sslobj.unwrap()
-            self.sslobj = None
-
-    def _read(self, n, initial=False):
-        """It seems that SSL Objects read() method may not supply as much
-        as you're asking for, at least with extremely large messages.
-        somewhere > 16K - found this in the test_channel.py test_large
-        unittest."""
-        result = ''
-
-        while len(result) < n:
+        if HAVE_PY26_SSL and self.sock is not None:
             try:
-                s = self.sslobj.read(n - len(result))
-            except socket.error, exc:
-                if not initial and exc.errno in (errno.EAGAIN, errno.EINTR):
+                unwrap = self.sock.unwrap
+            except AttributeError:
+                return
+            self.sock = unwrap()
+
+    def _read(self, n, initial=False,
+              _errnos=(errno.ENOENT, errno.EAGAIN, errno.EINTR)):
+        # According to SSL_read(3), it can at most return 16kb of data.
+        # Thus, we use an internal read buffer like TCPTransport._read
+        # to get the exact number of bytes wanted.
+        recv = self._quick_recv
+        rbuf = self._read_buffer
+        while len(rbuf) < n:
+            try:
+                s = recv(131072)  # see note above
+            except socket.error as exc:
+                # ssl.sock.read may cause ENOENT if the
+                # operation couldn't be performed (Issue celery#1414).
+                if not initial and exc.errno in _errnos:
                     continue
-                raise
+                raise exc
             if not s:
                 raise IOError('Socket closed')
-            result += s
+            rbuf += s
 
+        result, self._read_buffer = rbuf[:n], rbuf[n:]
         return result
 
     def _write(self, s):
         """Write a string out to the SSL socket fully."""
+        write = self.sock.write
         while s:
-            n = self.sslobj.write(s)
+            n = write(s)
             if not n:
                 raise IOError('Socket closed')
             s = s[n:]
@@ -222,24 +238,25 @@ class TCPTransport(_AbstractTransport):
         """Setup to _write() directly to the socket, and
         do our own buffered reads."""
         self._write = self.sock.sendall
-        self._read_buffer = bytes()
+        self._read_buffer = EMPTY_BUFFER
+        self._quick_recv = self.sock.recv
 
-    def _read(self, n, initial=False):
+    def _read(self, n, initial=False, _errnos=(errno.EAGAIN, errno.EINTR)):
         """Read exactly n bytes from the socket"""
-        while len(self._read_buffer) < n:
+        recv = self._quick_recv
+        rbuf = self._read_buffer
+        while len(rbuf) < n:
             try:
-                s = self.sock.recv(65536)
-            except socket.error, exc:
-                if not initial and exc.errno in (errno.EAGAIN, errno.EINTR):
+                s = recv(131072)
+            except socket.error as exc:
+                if not initial and exc.errno in _errnos:
                     continue
                 raise
             if not s:
                 raise IOError('Socket closed')
-            self._read_buffer += s
-
-        result = self._read_buffer[:n]
-        self._read_buffer = self._read_buffer[n:]
+            rbuf += s
 
+        result, self._read_buffer = rbuf[:n], rbuf[n:]
         return result
 
 
